import tensorflow as tf
from medgeconv.tf_ops import knn_graph


def get_knn_from_disjoint(nodes, k, is_valid):
    """
    Get the k nearest neighbors depending on distance.

    Parameters
    ----------
    nodes : tf.Tensor
        shape (None, n_features)
    k : int
        Number of nearest neighbors (excluding self).
    is_valid : tf.Tensor
        int32 tensor, shape (bs, n_points).

    Returns
    -------
    shape (None, k)
        int32, for each point, the indices of the points that are the
        nearest neighbors.

    """
    n_valid_nodes = tf.reduce_sum(is_valid, axis=-1)
    assert_op = tf.debugging.assert_greater(
        n_valid_nodes,
        k,
        message="one or more graphs in the batch had too few nodes for "
                "k nearest neighbors calculation!"
    )
    with tf.control_dependencies([assert_op]):
        indices, dists = knn_graph(nodes, n_valid_nodes, k+1)
    return indices[:, 1:]


def reduce_mean_valid_disjoint(nodes, is_valid):
    """
    Average over valid nodes.

    Parameters
    ----------
    nodes : tf.Tensor
        shape (None, n_features)
    is_valid : tf.Tensor
        int32 tensor shape (bs, n_points).

    Returns
    -------
    shape (bs, n_features)
        For each feature, aggeregated over all valid nodes.

    """
    graph_ids = get_graph_ids(is_valid)
    pooled = tf.math.segment_mean(nodes, graph_ids)
    pooled.set_shape(is_valid.shape[:1] + pooled.shape[1:])
    return pooled


def get_graph_ids(is_valid):
    """ Shape (None,). To which graph each node belongs to."""
    return tf.gather_nd(
        is_valid * tf.expand_dims(tf.range(tf.shape(is_valid)[0]), -1),
        tf.where(is_valid == 1))



import tensorflow as tf
from medgeconv.tf_ops import knn_graph


def get_knn(nodes, k):
    """
    Get the k nearest neighbors depending on distance.

    Parameters
    ----------
    nodes : tf.Tensor
        shape (None, n_features)
    k : int
        Number of nearest neighbors (excluding self).

    Returns
    -------
    shape (None, k)
        int32, for each point, the indices of the points that are the
        nearest neighbors.

    """
    n_valid_nodes = tf.cast(nodes.nested_row_lengths()[0], "int32")
    nodes_disjoint = nodes.merge_dims(0, 1)
    assert_op = tf.debugging.assert_greater(
        n_valid_nodes,
        k,
        message="one or more graphs in the batch had too few nodes for "
                "k nearest neighbors calculation!"
    )
    with tf.control_dependencies([assert_op]):
        indices, dists = knn_graph(nodes_disjoint, n_valid_nodes, k+1)
    return indices[:, 1:]


def get_xixj(nodes_disjoint, knn, k):
    """
    Get the features of each edge in the graph.

    Paramters
    ---------
    nodes : tf.Tensor
        shape (None, n_features)
    knn : tf.Tensor
        shape (None, k)
        int32, for each point, the indices of the points that are the
        nearest neighbors.
    k : int
        Number of nearest neighbors (excluding self).

    Returns
    -------
    tuple
        Two Tensors with shape (None, k, n_features).
        --> [?, i, j] desribes the edge between points i and j.
        The first matrix is xi, i.e. for each edge the central point
        The 2nd matrix is xj, i.e. for each edge the other point.

    """
    nodes_central = tf.tile(
        tf.expand_dims(nodes_disjoint, axis=-2),
        [1, k, 1]
    )

    # TODO this produces a 'Converting sparse IndexedSlices to a dense Tensor
    #  of unknown shape.' warning. Thats because nodes has an unknown shape
    #  (None, n_features), along first axis is gathered.
    nodes_neighbors = tf.gather(nodes_disjoint, knn)

    return nodes_central, nodes_neighbors



import tensorflow as tf
import tensorflow.keras as ks
import medgeconv.util as util
import medgeconv.layers as layers
import medgeconv.legacy.util as util_legacy


class DisjointEdgeConv:
    """
    The EdgeConv layer like it is used in ParticleNet, implemented
    as a pseudo-block and expecting disjoint input.
    Has a dense network as the kernel network.

    Input: (nodes, is_valid, coordinates)

    nodes, shape (batchsize, n_nodes_max, n_features)
        Node features of the graph, padded to fixed size.
    is_valid, shape (batchsize, n_nodes_max)
        1 for actual node, 0 for padded node.
    coordinates, shape (batchsize, n_nodes_max, n_coords)
        Features of each node used for calculating nearest
        neighbors.

    Output, shape (batchsize, n_nodes_max, units[-1])

    Parameters
    ----------
    units : List or int
        How many dense layers are in the kernel network, and how many
        neurons does each of them have? E.g. [64, 64] means two dense
        layers with 64 neurons each.
    next_neighbors : int
        How many next neighbors to construct the edge features with for
        each node.
    kernel_initializer : str
        For the kernel network.
    activation : str
        For the kernel network.
    shortcut : bool
        Add a shortcut connection between input and output?

    """
    def __init__(self, units,
                 next_neighbors=16,
                 kernel_initializer="glorot_uniform",
                 activation="relu",
                 shortcut=True):
        self.units = units
        self.next_neighbors = next_neighbors
        self.kernel_initializer = kernel_initializer
        self.activation = activation
        self.shortcut = shortcut

        self.kernel_network = lambda xi_xj: layers.kernel_nn(
            xi_xj, units=self.units,
            kernel_initializer=self.kernel_initializer,
            activation=self.activation)

    def __call__(self, inputs):
        nodes_disjoint, is_valid, coordinates_disjoint = inputs

        # get central and neighbour point features for each edge
        # between neighbours, defined by coordinates
        xi_xj = GetEdgeFeaturesDisjoint(
            next_neighbors=self.next_neighbors,
        )((nodes_disjoint, is_valid, coordinates_disjoint))

        x = self.get_edges(xi_xj)
        # apply MLP on points of each edge
        x = self.kernel_network(x)
        # aggregate over the nearest neighbours
        x = ks.backend.mean(x, axis=-2)

        if self.shortcut:
            sc = ks.layers.Dense(
                self.units[-1], use_bias=False,
                kernel_initializer=self.kernel_initializer)(nodes_disjoint)
            sc = ks.layers.BatchNormalization()(sc)
            x = ks.layers.Activation(self.activation)(sc + x)

        return x

    @staticmethod
    def get_edges(xi_xj):
        """
        Get the edge features, derived from the central point and
        the neighboring points.

        Output shape: (bs, n_nodes, n_neighbors, 2*n_features)

        """
        xi, xj = xi_xj
        dif = ks.layers.Subtract()([xi, xj])
        x = ks.layers.Concatenate(axis=-1)([xi, dif])
        return x


class GetEdgeFeaturesDisjoint(ks.layers.Layer):
    """
    Get the edge features of the graph.

    Input shape:
        [(None, n_features),
         (bs, n_nodes),
         (None, n_coordinates)]
        --> [nodes, is_valid, coordinates]
        - all the nodes of the graph, together with their features
        - Info about which nodes are valid nodes
        - Coordinates that will be used for calculateing the distance / knn

    Output shape:
        [
         (None, next_neighbors, n_features),
         (None, next_neighbors, n_features),
        ]
        --> [xi, xj]
        These matrices represent the edges of the graph:
        for [, i, j]:
        matrix 1 is xi (the central point),
        and matrix 2 is xj (the nearest neighbor j of central point i)

    """
    def __init__(self, next_neighbors, **kwargs):
        super().__init__(**kwargs)
        self.next_neighbors = next_neighbors

    def call(self, inputs):
        nodes_disjoint, is_valid, coordinates_disjoint = inputs
        # get the k nearest neighbours for each node
        knn_disjoint = util_legacy.get_knn_from_disjoint(
            coordinates_disjoint,
            self.next_neighbors,
            is_valid=is_valid,
        )
        # get central and neighbour point for each edge between neighbours
        return util.get_xixj(
            nodes_disjoint, knn_disjoint, k=self.next_neighbors)

    def compute_output_shape(self, input_shape):
        return (input_shape[0, :1] + (self.next_neighbors,) + input_shape[0, -1:], ) * 2

    def get_config(self):
        config = super().get_config()
        config["next_neighbors"] = self.next_neighbors
        return config


class DenseToDisjoint(ks.layers.Layer):
    """ (batchsize, n_nodes, n_features) --> (None, n_features) """
    def call(self, inputs):
        nodes, is_valid, coordinates = inputs
        is_valid = tf.cast(is_valid, "int32")

        valid_indices = tf.where(is_valid == 1)
        nodes_disjoint = tf.gather_nd(nodes, valid_indices)
        coordinates_disjoint = tf.gather_nd(coordinates, valid_indices)

        return nodes_disjoint, is_valid, coordinates_disjoint


class GlobalAvgPoolingDisjoint(ks.layers.Layer):
    """
    An average valid pooling layer.
    Pools a graph by averaging features over all nodes (or only all
    valid nodes).

    Input shape:    [(None, n_features), (bs, n_nodes)]
    Output shape:   (bs, n_features)

    """
    def call(self, inputs):
        nodes_disjoint, is_valid = inputs
        return util_legacy.reduce_mean_valid_disjoint(
            nodes_disjoint, is_valid)

    def compute_output_shape(self, input_shape):
        return input_shape[1][:1] + input_shape[0][-1:]



import tensorflow as tf
from tensorflow.python.framework import load_library
from tensorflow.python.platform import resource_loader
from tensorflow.python.framework import ops

_knn_op_module = load_library.load_op_library(
    resource_loader.get_path_to_datafile('./_knn_graph_ops.so'))
_knn_graph_op = _knn_op_module.knn_graph


ops.NotDifferentiable("KnnGraph")


def knn_graph(nodes_disjoint, n_nodes_per_graph, k):
    """
    k-nearest neighbors on a batch of graphs.

    Parameters
    ----------
    nodes_disjoint : tf.Tensor
        The disjoint node features. shape (n_nodes, n_dims), float32.
    n_nodes_per_graph : tf.Tensor
        The number of nodes per graph. Shape (batchsize, ), int32.
        E.g., if there are 3 graphs in the batch with 3, 5 and 4 nodes,
        n_nodes_per_graph is [3, 5, 4].
    k : int
        The number of nearest neighbors, including self connection.
        I.e. k=16 means self and 15 neighbors.

    Returns
    -------
    indices : tf.Tensor
        The nearest neighbor indices. Shape (n_nodes, k), int32.
    dists : tf.Tensor
        The distances to the neighbors. Shape (n_nodes, k), float32.

    """
    x_ptr = tf.concat([[0, ], tf.math.cumsum(n_nodes_per_graph)], axis=0)
    indices, dists = _knn_graph_op(nodes_disjoint, x_ptr, k)
    return indices, dists



"""
TF 2.3 (GTX 1080)
-----------------
Total time: 23.3371 s
Per batch: 235.7 ms
   median: 236.9 ms
Per sample: 3.6833 ms
   median: 3.7010 ms

TF 2.5 (GTX 1080)
-----------------
Total time: 23.7269 s
Per batch: 239.7 ms
   median: 238.4 ms
Per sample: 3.7448 ms
   median: 3.7243 ms

TF 2.5 & MEdgeconv 1.2 (ragged) (GTX 1080)
-----------------
Total time: 19.8413 s
Per batch: 200.4 ms
   median: 197.4 ms
Per sample: 3.1315 ms
   median: 3.0845 ms

"""

import time
import numpy as np
import tensorflow as tf
from medgeconv import DisjointEdgeConvBlock


NEXT_NEIGHBORS = 16
N_POINTS_MAX = 2000
N_FEATURES = 7
N_COORDS = 4
BATCHSIZE = 64
UNITS = (64, 64, 64)
N_NODES_OUTPUT = 2


def get_model():
    inps = (
        tf.keras.layers.Input((None, N_FEATURES), ragged=True),
        tf.keras.layers.Input((None, N_COORDS), ragged=True)
    )
    x = DisjointEdgeConvBlock(
        units=UNITS,
        next_neighbors=NEXT_NEIGHBORS,
        kernel_initializer="ones",
        pooling=True,
    )(inps)
    x = tf.keras.layers.Dense(N_NODES_OUTPUT)(x)
    model = tf.keras.Model(inps, x)
    model.compile("sgd", "mse")
    return model


class DataGenerator(tf.keras.utils.Sequence):
    def __init__(self, length=1000, batches=10):
        self.length = length
        self.batches = batches
        self.x, self.y = get_data(self.batches)

    def __len__(self):
        return self.length

    def __getitem__(self, item):
        item = item % self.batches
        slc = slice(item*BATCHSIZE, (item+1)*BATCHSIZE)
        return [x_[slc] for x_ in self.x], self.y[slc]


def get_data(batches=10):
    """ Get all batches of data. Number of valid nodes is varied per graph. """
    # number of valid nodes, linearily decreasing
    n_valid_nodes = np.random.triangular(
        NEXT_NEIGHBORS+1, NEXT_NEIGHBORS+1, N_POINTS_MAX+1, batches*BATCHSIZE).astype(int)
    # exponential
    # np.minimum(N_POINTS_MAX+1, np.random.exponential(300, batches*BATCHSIZE)).astype(int) + NEXT_NEIGHBORS+1
    data = np.random.rand(n_valid_nodes.sum(), N_FEATURES).astype("float32")
    nodes = tf.RaggedTensor.from_row_lengths(data, n_valid_nodes)
    coords = nodes[:, :, :N_COORDS]

    x = nodes, coords
    y = np.zeros((batches*BATCHSIZE, N_NODES_OUTPUT))
    return x, y


def benchmark(warmup_batches=10, profile_batches=10, n_loops=100, tensorboarding=False):
    """
    Show average fit times of a simple model.

    For tensorboard profiling on GPU:
        (0.)  pip install -U tensorboard_plugin_profile
         1.   module load cuda/10.1
         2.   export LD_LIBRARY_PATH=/usr/local/cuda-10.1/extras/CUPTI/lib64/

    To display:
        1. firefox &
        2. tensorboard --logdir benchmarks/
        3. URL:  http://localhost:6006/

    """
    model = get_model()
    model.summary()

    if tensorboarding:
        benchmark_dir = "benchmarks"
        print(f"TensorBoarding to '{benchmark_dir}'")
        tensorboard = tf.keras.callbacks.TensorBoard(
            log_dir=benchmark_dir,
            profile_batch=f"{warmup_batches+1}, {warmup_batches+profile_batches}",
            histogram_freq=1,
            write_images=True,
        )
        generator = DataGenerator(length=warmup_batches+profile_batches+1)
        val_data = generator[warmup_batches+profile_batches+1]
        model.fit(
            x=generator,
            validation_data=val_data,
            verbose=1,
            callbacks=[tensorboard],
            steps_per_epoch=warmup_batches+profile_batches,
        )
        print("tensorboarding complete")

    times = np.zeros(n_loops)
    generator = DataGenerator(length=n_loops)
    for i in range(n_loops):
        if i > 0 and i % 10 == 0:
            print(f"Batch {i}/{n_loops}...\t({times[i-10:i].mean():.4f} s per batch)")
        x, y = generator[i]
        start_time = time.time()
        model.fit(x, y, verbose=0)
        times[i] = time.time() - start_time
    print("\nDone! Times excluding first batch:")
    print(f"Total time: {times[1:].sum():.4f} s")
    print(f"Per batch: {1000*times[1:].mean():.1f} ms")
    print(f"   median: {1000*np.median(times[1:]):.1f} ms")
    print(f"Per sample: {1000*times[1:].mean()/BATCHSIZE:.4f} ms")
    print(f"   median: {1000*np.median(times[1:])/BATCHSIZE:.4f} ms")
    return times


if __name__ == '__main__':
    benchmark()




import tensorflow as tf
import medgeconv.layers as layers
import medgeconv.legacy.layers as layers_legacy


class DisjointEdgeConvBlock:
    """
    EdgeConv with additional options as used in the ParticleNet paper.

    Parameters
    ----------
    units : List or int
        How many dense layers are in the kernel network, and how many
        neurons does each of them have? E.g. [64, 64] means two dense
        layers with 64 neurons each.
    next_neighbors : int
        How many next neighbors to construct the edge features with for
        each node.
    kernel_initializer : str
        For the kernel network.
    activation : str
        For the kernel network.
    shortcut : bool
        Add a shortcut connection between input and output?
    batchnorm_for_nodes : bool
        Use batchnorm layers for node features?
        Recommended for first layer.
    pooling : bool
        Finish with average pooling?
        Has to be used if this is the last layer.

    """
    def __init__(self, units,
                 next_neighbors,
                 kernel_initializer="glorot_uniform",
                 activation="relu",
                 shortcut=True,
                 batchnorm_for_nodes=False,
                 pooling=False):
        self.batchnorm_for_nodes = batchnorm_for_nodes
        self.pooling = pooling

        self.edgeconv = layers.DisjointEdgeConv(
            units=units,
            next_neighbors=next_neighbors,
            kernel_initializer=kernel_initializer,
            activation=activation,
            shortcut=shortcut,
        )

    def __call__(self, x):
        nodes, coordinates = x

        if self.batchnorm_for_nodes:
            nodes_disjoint = nodes.merge_dims(0, 1)
            nodes_disjoint = tf.keras.layers.BatchNormalization()(nodes_disjoint)
            nodes = tf.RaggedTensor.from_row_splits(nodes_disjoint, nodes.row_splits)

        nodes = self.edgeconv((nodes, coordinates))

        if self.pooling:
            return tf.keras.layers.GlobalAvgPool1D()(nodes)
        else:
            return nodes, nodes


custom_objects = {
    "GetEdgeFeatures": layers.GetEdgeFeatures,
    # legacy:
    "GlobalAvgPoolingDisjoint": layers_legacy.GlobalAvgPoolingDisjoint,
    "GetEdgeFeaturesDisjoint": layers_legacy.GetEdgeFeaturesDisjoint,
    "DenseToDisjoint": layers_legacy.DenseToDisjoint,
}



import tensorflow as tf
from medgeconv import util


class TestTFFunctions(tf.test.TestCase):
    def test_get_knn_from_points(self):
        points = tf.constant([
            [0, 0], [0, 3], [0, 2],
            [1, 1], [4, 5], [5, 5], [1, 1.1],
        ], dtype="float32")
        n_nodes = tf.constant([3, 4], dtype="int32")
        points = tf.RaggedTensor.from_row_lengths(points, n_nodes)

        target = tf.constant([
            [2, 1], [2, 0], [1, 0],
            [6, 4], [5, 6], [4, 6], [3, 4],
        ], dtype="int32")
        result = util.get_knn(points, k=2)
        self.assertAllEqual(result, target)

    def test_get_knn_from_points_too_few_neighbors(self):
        points = tf.constant([
            [0, 0], [0, 3],
            [1, 1], [4, 5], [5, 5], [1, 1.1],
        ], dtype="float32")
        n_nodes = tf.constant([2, 4], dtype="int32")
        points = tf.RaggedTensor.from_row_lengths(points, n_nodes)

        with self.assertRaises(tf.errors.InvalidArgumentError):
            util.get_knn(points, k=2)

    def test_get_knn_from_points_eager(self):
        """ map_fn turns the func into a tf.function apparently """
        tf.config.run_functions_eagerly(True)
        try:
            self.test_get_knn_from_points()
        finally:
            tf.config.run_functions_eagerly(False)



import os
import tempfile
import tensorflow as tf
import numpy as np
import medgeconv


class TestEdgy(tf.test.TestCase):
    def setUp(self):
        self.n_points, self.n_features, self.n_coords = 5, 4, 3
        self.batchsize = 2
        self.units = [5, 2]
        self.next_neighbors = 3
        self.n_valid_nodes = [] * self.batchsize

        inp_points = tf.keras.layers.Input(
            (None, self.n_features), ragged=True)
        inp_coords = tf.keras.layers.Input(
            (None, self.n_coords), ragged=True)
        self.inps = (inp_points, inp_coords)
        self.edge_layer = medgeconv.DisjointEdgeConvBlock(
            units=self.units,
            next_neighbors=self.next_neighbors,
            kernel_initializer="ones",
            pooling=True,
        )
        self.edge_out = self.edge_layer(self.inps)
        self.model = tf.keras.Model(self.inps, self.edge_out)

        points = np.arange(
            self.batchsize * self.n_points * self.n_features).reshape(
            (self.batchsize, self.n_points, self.n_features)).astype("float32")
        points = tf.RaggedTensor.from_tensor(points[:, :-1])
        coords = points[:, :, :self.n_coords]
        self.x = points, coords
        self.y = np.zeros((self.batchsize, self.units[-1]))

    def test_output_shape(self):
        self.assertTupleEqual(
            self.model.output_shape, (None, self.units[-1]))

    def test_output(self):
        output = self.model.predict(self.x)
        target = np.array([
            [216.46521, 216.46521],
            [659.3956, 659.3956]], dtype="float32")
        np.testing.assert_almost_equal(target, output, decimal=4)

    def test_train(self):
        self.model.compile("sgd", "mse")
        train_loss = self.model.train_on_batch(x=self.x, y=self.y)
        np.testing.assert_almost_equal(train_loss, 1.4926, decimal=4)

    def test_loading(self):
        self.model.compile("sgd", "mse")
        with tempfile.TemporaryDirectory() as tempdir:
            path = os.path.join(tempdir, "temp.h5")
            tf.keras.models.save_model(self.model, path)
            loaded = tf.keras.models.load_model(
                path, custom_objects=medgeconv.custom_objects)
            train_loss = loaded.train_on_batch(x=self.x, y=self.y)
            np.testing.assert_almost_equal(train_loss, 1.4926, decimal=4)


class RepeatInEager(TestEdgy):
    """ Repeat the tests above in eager mode. Mostly to get the coverage right.
    """
    def setUp(self):
        tf.config.run_functions_eagerly(True)
        super().setUp()
    
    def tearDown(self):
        tf.config.run_functions_eagerly(False)



#!/usr/bin/env python
from setuptools import setup, find_packages

with open('requirements.txt') as fobj:
    requirements = [l.strip() for l in fobj.readlines()]

with open("README.rst") as fh:
    long_description = fh.read()

setup(
    name='medgeconv',
    author='Stefan Reck',
    author_email='stefan.reck@fau.de',
    long_description=long_description,
    setup_requires=['setuptools_scm'],
    use_scm_version={
        "local_scheme": "no-local-version",
    },
    install_requires=requirements,
    packages=find_packages(exclude=("tests", )),
    include_package_data=True,
    license='MIT',
)
